' Copyright (c) 2013 Ivan Krivyakov
' Licensed under the Apache License, Version 2.0 (the "License")
' you may not use this file except in compliance with the License.
' You may obtain a copy of the License at
' http://www.apache.org/licenses/LICENSE-2.0
'

Namespace BitnessCheck
	''' <summary>
	''' Interaction logic for BitnessCheck.xaml
	''' </summary>
	Partial Public Class MainUserControl
		Inherits UserControl

		Private _physicalMemory As Integer
		Private _virtualMemory As Integer
		Private _memory As New List(Of Byte)

		Public Sub New()
			InitializeComponent()
			BitnessText.Text = "Running at " & IntPtr.Size*8 & " bit"
			UpdateMemoryText()
		End Sub

		Private Sub UpdateMemoryText()
			PhysicalMemoryText.Text = _physicalMemory & " G"
			VirtualMemoryText.Text = _virtualMemory & " G"
		End Sub

		Private Sub Reserve_Click(ByVal sender As Object, ByVal e As RoutedEventArgs)
			Allocate(AddressOf Reserve)
		End Sub

		Private Sub Commit_Click(ByVal sender As Object, ByVal e As RoutedEventArgs)
			Allocate(AddressOf Commit)
		End Sub

'         To outstmart .NET optimization and virtual memory mechanism we need to keep in mind these facts
'         *
'         * 1. It is not possible to allocate more than 2G of memory at a time, even when running as 64 bit
'         *
'         * 2. If we allocate memory in several independent arrays, they might be immediately garbage collected.
'         *    We must reference all arrays from a single data structure, so it can only be garbage collected as a whole.
'         * 
'         * 3. To have virtual address space actually reserved, we must access at least one byte of the memory allocated.
'         *    32-bit process cannot have 3G of virtual address space reserved and exception will be thrown at this point.
'         *    
'         * 4. If we want pages of physical memory allocated, and not just virtual address space reserved, we must explicitly
'         *    access each page, which happens to be 4096 bytes on Intel systems. This is what Commit() method does.
'         
		Private Sub Allocate(ByVal commitAction As Action(Of Byte()))
			Try
				Const oneG As Long = 1024L * 1024 * 1024
				Dim memory(oneG - 1) As Byte
                commitAction(memory)

                For Each m1 As Byte In memory
                    _memory.Add(m1)
                Next

			Catch ex As Exception
				MessageBox.Show(ex.ToString, "Failure")
			End Try

			UpdateMemoryText()
		End Sub

		Private Sub Reserve(ByVal b() As Byte)
			b(0) = 42
            _virtualMemory += 1
		End Sub

		Private Sub Commit(ByVal b() As Byte)
			Const PageSize As Integer = 4096

			' index can be int, as currently array size cannot exceed 2G
            For index  = 0 To b.LongLength - 1 Step PageSize
                b(index) = 42 ' actual value does not matter
            Next index

            _virtualMemory += 1
            _physicalMemory += 1
		End Sub
	End Class
End Namespace
