' Copyright 2009 Ivan Krivyakov
'
'   Licensed under the Apache License, Version 2.0 (the "License")
'   you may not use this file except in compliance with the License.
'   You may obtain a copy of the License at
'
'       http://www.apache.org/licenses/LICENSE-2.0
'
'   Unless required by applicable law or agreed to in writing, software
'   distributed under the License is distributed on an "AS IS" BASIS,
'   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
'   See the License for the specific language governing permissions and
'   limitations under the License.
' 

Imports System
Imports System.ComponentModel
Imports System.Windows.Threading

Namespace SolarSystem
	Friend Class OrbitsCalculator
		Implements INotifyPropertyChanged

		Private _startTime As Date
		Private _startDays As Double
		Private _timer As DispatcherTimer

		Private Const EarthYear As Double = 365.25
		Private Const EarthRotationPeriod As Double = 1.0
		Private Const SunRotationPeriod As Double = 25.0
		Private Const TwoPi As Double = Math.PI * 2

		Private _daysPerSecond As Double = 2
		Public Property DaysPerSecond As Double
			Get
				Return _daysPerSecond
			End Get
			Set(ByVal value As Double)
				_daysPerSecond = value
				Update("DaysPerSecond")
			End Set
		End Property

		Public Property EarthOrbitRadius As Double
			Get
				Return 40
			End Get
			Set(ByVal value As Double)
			End Set
		End Property

		Public Property Days As Double

		Public Property EarthRotationAngle As Double

		Public Property SunRotationAngle As Double

		Public Property EarthOrbitPositionX As Double

		Public Property EarthOrbitPositionY As Double

		Public Property EarthOrbitPositionZ As Double

		Public Property ReverseTime As Boolean

		Public Property Paused As Boolean

		Public Sub New()
			EarthOrbitPositionX = EarthOrbitRadius
			DaysPerSecond = 2
		End Sub

		Public Sub StartTimer()
			_startTime = Date.Now
			_timer = New DispatcherTimer
			_timer.Interval = TimeSpan.FromMilliseconds(100)
			AddHandler _timer.Tick, AddressOf OnTimerTick
			_timer.Start()
		End Sub

		Private Sub StopTimer()
			_timer.Stop()
			RemoveHandler _timer.Tick, AddressOf OnTimerTick
			_timer = Nothing
		End Sub

		Public Sub Pause(ByVal doPause As Boolean)
			If doPause Then
				StopTimer()
			Else
				StartTimer()
			End If
		End Sub

		Private Sub OnTimerTick(ByVal sender As Object, ByVal e As EventArgs)
			Dim now = Date.Now
			Days += (now-_startTime).TotalMilliseconds * DaysPerSecond / 1000.0 * (If(ReverseTime, -1, 1))
			_startTime = now
			Update("Days")
			OnTimeChanged()
		End Sub

		Private Sub OnTimeChanged()
			EarthPosition()
			EarthRotation()
			SunRotation()
		End Sub

		Private Sub EarthPosition()
			Dim angle As Double = 2 * Math.PI * Days / EarthYear
			EarthOrbitPositionX = EarthOrbitRadius * Math.Cos(angle)
			EarthOrbitPositionY = EarthOrbitRadius * Math.Sin(angle)
			Update("EarthOrbitPositionX")
			Update("EarthOrbitPositionY")
		End Sub

		Private Sub EarthRotation()
			EarthRotationAngle = 360 * Days / EarthRotationPeriod
			Update("EarthRotationAngle")
		End Sub

		Private Sub SunRotation()
			SunRotationAngle = 360 * Days / SunRotationPeriod
			Update("SunRotationAngle")
		End Sub

        Private Sub Update(ByVal propertyName As String)
            'If PropertyChanged IsNot Nothing Then
            Dim args = New PropertyChangedEventArgs(propertyName)
            RaiseEvent PropertyChanged(Me, args)
            'End If
        End Sub

        Public Event PropertyChanged As PropertyChangedEventHandler Implements INotifyPropertyChanged.PropertyChanged
	End Class
End Namespace
