' Copyright (c) 2013 Ivan Krivyakov
' Licensed under the Apache License, Version 2.0 (the "License")
' you may not use this file except in compliance with the License.
' You may obtain a copy of the License at
' http://www.apache.org/licenses/LICENSE-2.0
'

Imports System.Reflection
Imports System.Runtime.Remoting
Imports System.Threading
Imports System.Windows.Threading
Imports IKriv.PluginHosting.Log
Imports IKriv.WpfHost.Interfaces

Namespace IKriv.PluginHosting
	''' <summary>
	''' Loads plugins for the host
	''' </summary>
	Public Class PluginLoader
		Inherits MarshalByRefObject
		Implements IPluginLoader

		Private _dispatcher As Dispatcher
		Private _log As LocalLog
		Private _host As IWpfHost
		Private _name As String

		Public Sub Run(ByVal name As String)
			_name = name
			_dispatcher = Dispatcher.CurrentDispatcher

			_log = CreateLog()
			Using _log
				Try
                    _log.Info("PluginHost running at " & IntPtr.Size * 8 & " bit, CLR version " & Environment.Version.ToString)
					CType(New AssemblyResolver, AssemblyResolver).Setup
					AddHandler AppDomain.CurrentDomain.UnhandledException, AddressOf OnUnhandledException
					IpcServices.RegisterChannel(name)
					RegisterObject()
					SignalReady()
					Dispatcher.Run()
				Catch ex As Exception
					ReportFatalError(ex)
				End Try

				Thread.Sleep(100) ' allow any pending remoting operations to finish
				_log.Debug("Shutdown complete")
			End Using
		End Sub

		Public Function LoadPlugin(ByVal host As IWpfHost, ByVal startupInfo As PluginStartupInfo) As IRemotePlugin Implements IPluginLoader.LoadPlugin
			_host = host

			_log.Info(String.Format("LoadPlugin('{0}','{1}')", startupInfo.AssemblyName, startupInfo.MainClass))

			CType(New ProcessMonitor(AddressOf Dispose), ProcessMonitor).Start(_host.HostProcessId)

			Dim createOnUiThread As Func(Of PluginStartupInfo, Object) = AddressOf LoadPluginOnUiThread
			Dim result = _dispatcher.Invoke(createOnUiThread, startupInfo)

			_log.Debug("Returning plugin object to host")

			If TypeOf result Is Exception Then
				_log.Error("Error loading plugin", CType(result, Exception))
				Throw New TargetInvocationException(CType(result, Exception))
			End If
			Return CType(result, IRemotePlugin)
		End Function

		Public Sub Dispose() Implements IDisposable.Dispose
			_log.Info("Shutdown requested")

			If _dispatcher IsNot Nothing Then
				_log.Debug("Performing dispatcher shutdown")
				_dispatcher.BeginInvokeShutdown(DispatcherPriority.Normal)
			Else
				_log.Debug("No dispatcher, exiting the process")
				_log.Dispose()
				Environment.Exit(1)
			End If
		End Sub

		Public Overrides Function InitializeLifetimeService() As Object
			Return Nothing ' live forever
		End Function

		Private Function LoadPluginOnUiThread(ByVal startupInfo As PluginStartupInfo) As Object
			_log.Debug("Creating plugin on UI thread")

			Dim assembly = startupInfo.AssemblyName
			Dim mainClass = startupInfo.MainClass

			Try
				Dim obj = PluginCreator.CreatePlugin(startupInfo.AssemblyName, startupInfo.MainClass, _host)
				_log.Debug("Created local plugin class instance")

				Dim localPlugin = TryCast(obj, IPlugin)

				If localPlugin Is Nothing Then
					Dim message = String.Format("Object of type {0} cannot be loaded as plugin " & "because it does not implement IPlugin interface", mainClass)

					Throw New InvalidOperationException(message)
				End If

				Dim remotePlugin = New RemotePlugin(localPlugin)
				_log.Debug("Created plugin control")
				Return remotePlugin
			Catch ex As Exception
				Dim message = String.Format("Error loading type '{0}' from assembly '{1}'. {2}", mainClass, assembly, ex.Message)

				Return New ApplicationException(message, ex)
			End Try
		End Function

		Private Sub RegisterObject()
			_log.Debug(String.Format("Listening on ipc://{0}/PluginLoader", _name))
			RemotingServices.Marshal(Me, "PluginLoader", GetType(IPluginLoader))
		End Sub

		Private Sub SignalReady()
			Dim eventName = _name & ".Ready"
			Dim readyEvent = EventWaitHandle.OpenExisting(eventName)
			readyEvent.Set()
		End Sub

		Private Sub OnUnhandledException(ByVal sender As Object, ByVal e As UnhandledExceptionEventArgs)
			Dim exception = If((TryCast(e.ExceptionObject, Exception)), New Exception("Unknown error. Exception object is null"))

			ReportFatalError(exception)
		End Sub

		Private Sub ReportFatalError(ByVal exception As Exception)
			_log.Error("Unhandled exception", exception)

			If _host IsNot Nothing Then
				_log.Debug("Reporting fatal error to host")
				_host.ReportFatalError(ExceptionUtil.GetUserMessage(exception), exception.ToString)
			Else
				_log.Warn("Host is null, cannot report error to host")
			End If

			_log.Info("Exiting the process to prevent 'program stopped working' dialog")
			_log.Dispose() ' flush pending data
			Environment.Exit(2)
		End Sub

        Private Function CreateLog() As LocalLog
            Return New LocalLog().Console(LogLevel.Debug).Trace(_name, LogLevel.Debug).File(_name, LogLevel.Debug)
        End Function
	End Class
End Namespace
