' Copyright (c) 2013 Ivan Krivyakov
' Licensed under the Apache License, Version 2.0 (the "License")
' you may not use this file except in compliance with the License.
' You may obtain a copy of the License at
' http://www.apache.org/licenses/LICENSE-2.0
'

Imports System.Collections.ObjectModel
Imports System.Text
Imports System.Threading.Tasks
Imports IKriv.WpfHost.Mvvm

Namespace IKriv.WpfHost
    Friend Class MainViewModel
        Inherits ViewModelBase
        Implements IDisposable

        Private ReadOnly _pluginController As PluginController
        Private ReadOnly _errorHandlingService As ErrorHandlingService

        Public Sub New(ByVal errorHandlingService As ErrorHandlingService, ByVal pluginController As PluginController)
            _errorHandlingService = errorHandlingService
            _pluginController = pluginController
            LoadPluginCommand = New DelegateCommand(Of PluginCatalogEntry)(AddressOf LoadPlugin)
            CloseTabCommand = New DelegateCommand(Of Plugin)(AddressOf CloseTab)

            _pluginController.LoadCatalogAcync.ContinueWith(Sub(unusedTask) AvailablePlugins = _pluginController.AvailablePlugins)
        End Sub

        Private _availablePlugins As IEnumerable(Of PluginCatalogEntry)
        Public Property AvailablePlugins As IEnumerable(Of PluginCatalogEntry)
            Get
                Return _availablePlugins
            End Get
            Set(value As IEnumerable(Of PluginCatalogEntry))
                _availablePlugins = value
                RaisePropertyChanged("AvailablePlugins")
            End Set
        End Property

        Public ReadOnly Property LoadedPlugins As ObservableCollection(Of Plugin)
            Get
                Return _pluginController.LoadedPlugins
            End Get
        End Property

        Private _selectedPlugin As Plugin
        Public Property SelectedPlugin As Plugin
            Get
                Return _selectedPlugin
            End Get
            Set(value As Plugin)
                _selectedPlugin = value
                RaisePropertyChanged("SelectedPlugin")
            End Set
        End Property

        Private PrivateLoadPluginCommand As ICommand
        Public Property LoadPluginCommand As ICommand
            Get
                Return PrivateLoadPluginCommand
            End Get
            Private Set(ByVal value As ICommand)
                PrivateLoadPluginCommand = value
            End Set
        End Property

        Private PrivateCloseTabCommand As ICommand
        Public Property CloseTabCommand As ICommand
            Get
                Return PrivateCloseTabCommand
            End Get
            Private Set(ByVal value As ICommand)
                PrivateCloseTabCommand = value
            End Set
        End Property

        Public Sub Dispose() Implements IDisposable.Dispose
            _pluginController.Dispose()
        End Sub

        Public Function CanClose() As Boolean
            Dim unsavedItems = _pluginController.GetUnsavedItems
            If unsavedItems.Count = 0 Then
                Return True
            End If

            Dim sb = New StringBuilder

            sb.AppendLine("The following items are not saved:")
            For Each unsavedItem In unsavedItems
                sb.AppendLine(vbTab & unsavedItem.Key.Title & ":")
                For Each item In unsavedItem.Value
                    sb.AppendLine(vbTab & vbTab & item)
                Next item
            Next unsavedItem

            sb.AppendLine()
            sb.Append("Are you sure you want to close the application and lose this data?")
            Return _errorHandlingService.Confirm(sb.ToString)
        End Function

#If CSHARP_45 Then
        Private Async Sub LoadPlugin(ByVal catalogEntry As PluginCatalogEntry)
            Try
                Dim plugin = Await _pluginController.LoadPluginAsync(catalogEntry)
                SelectedPlugin = plugin
            Catch ex As Exception
                _errorHandlingService.ShowError("Error loading plugin", ex)
            End Try
        End Sub

#Else

		Private Sub LoadPlugin(ByVal catalogEntry As PluginCatalogEntry)
			Dim task = _pluginController.LoadPluginAsync(catalogEntry)
			Dim scheduler = TaskScheduler.FromCurrentSynchronizationContext
			task.ContinueWith(Sub(t)
				Try
					SelectedPlugin = t.Result
				Catch ex As Exception
					_errorHandlingService.ShowError("Error loading plugin", ex)
				End Try
			End Sub, scheduler)
		End Sub
#End If

        Private Sub CloseTab(ByVal plugin As Plugin)
            If Not CanClose(plugin) Then
                Exit Sub
            End If

            Dim changeSelection As Boolean = (plugin Is SelectedPlugin)
            Dim selectedIndex As Integer = LoadedPlugins.IndexOf(plugin)

            _pluginController.RemovePlugin(plugin)

            If changeSelection Then
                Dim count As Integer = LoadedPlugins.Count

                If count = 0 Then
                    SelectedPlugin = Nothing
                Else
                    If selectedIndex >= count Then
                        selectedIndex = count - 1
                    End If
                    SelectedPlugin = LoadedPlugins(selectedIndex)
                End If
            End If
        End Sub

        Private Function CanClose(ByVal plugin As Plugin) As Boolean
            Dim unsavedItems = _pluginController.GetUnsavedItems(plugin)
            If unsavedItems Is Nothing OrElse unsavedItems.Length = 0 Then
                Return True
            End If

            Dim message = "The following items are not saved:" & vbCrLf & String.Join(vbCrLf, unsavedItems) & vbCrLf & vbCrLf & "Are you sure you want to close " & plugin.Title & "?"

            Return _errorHandlingService.Confirm(message)
        End Function
    End Class
End Namespace

