' Copyright (c) 2013 Ivan Krivyakov
' Licensed under the Apache License, Version 2.0 (the "License")
' you may not use this file except in compliance with the License.
' You may obtain a copy of the License at
' http://www.apache.org/licenses/LICENSE-2.0
'

Imports System.AddIn.Pipeline
Imports System.IO
Imports IKriv.PluginHosting
Imports IKriv.WpfHost.Interfaces
Imports Microsoft.Practices.Unity

Namespace IKriv.WpfHost
	Friend Class Plugin
		Implements IServiceProvider, IDisposable

		Private ReadOnly _childContainer As IUnityContainer
		Private ReadOnly _log As ILog
		Private _remoteProcess As PluginProcessProxy
		Private _startupInfp As PluginStartupInfo
		Private _isDisposing As Boolean
		Private _fatalErrorOccurred As Boolean

        Private PrivateCatalogEntry As PluginCatalogEntry
		Public Property CatalogEntry As PluginCatalogEntry
			Get
                Return PrivateCatalogEntry
			End Get
			Private Set(ByVal value As PluginCatalogEntry)
                PrivateCatalogEntry = value
			End Set
		End Property

        Private PrivateView As FrameworkElement
		Public Property View As FrameworkElement
			Get
                Return PrivateView
			End Get
			Private Set(ByVal value As FrameworkElement)
                PrivateView = value
			End Set
		End Property

        Private PrivateTitle As String
		Public Property Title As String
			Get
                Return PrivateTitle
			End Get
			Private Set(ByVal value As String)
                PrivateTitle = value
			End Set
		End Property

		Public Event [Error] As EventHandler(Of PluginErrorEventArgs)

		Public Sub New(ByVal container As IUnityContainer, ByVal log As ILog)
			_log = log
			_childContainer = container.CreateChildContainer
		End Sub

		' can be executed on any thread
		Public Sub Load(ByVal catalogEntry As PluginCatalogEntry)
			If Me.CatalogEntry IsNot Nothing Then
				Throw New InvalidOperationException("Plugin can be loaded only once")
			End If

			Me.CatalogEntry = catalogEntry
			Title = catalogEntry.Name

			Initialize()

			_log.Debug(String.Format("Loading plugin {0} from {1}, {2}", Me.CatalogEntry.Name, Me.CatalogEntry.AssemblyPath, Me.CatalogEntry.MainClass))

			Dim host = _childContainer.Resolve(Of PluginViewOfHost)()
			AddHandler host.FatalError, AddressOf OnFatalError

			_remoteProcess = _childContainer.Resolve(Of PluginProcessProxy)()

			_log.Debug("Starting plugin process")
			_remoteProcess.Start()
			CType(New ProcessMonitor(AddressOf OnProcessExited), ProcessMonitor).Start(_remoteProcess.Process)

			_log.Debug("Calling LoadPlugin()")
			_remoteProcess.LoadPlugin()
		End Sub

		' must execute on UI thread
		Public Sub CreateView()
			_log.Debug("Creating plugin view")
			View = FrameworkElementAdapters.ContractToViewAdapter(_remoteProcess.RemotePlugin.Contract)
			_log.Debug("Plugin view created")
		End Sub

		Public Function GetService(ByVal serviceType As Type) As Object Implements IServiceProvider.GetService
			If _remoteProcess Is Nothing Then
				Return Nothing
			End If
			Return _remoteProcess.RemotePlugin.GetService(serviceType)
		End Function

		Public Sub Dispose() Implements IDisposable.Dispose
			_isDisposing = True

			Try
				Dim disposableView = TryCast(View, IDisposable)
				If disposableView IsNot Nothing Then
					disposableView.Dispose()
				End If
			Catch ex As Exception
				ReportError("Error when disposing view", ex)
			End Try

			_childContainer.Dispose()
		End Sub

		Private Sub Initialize()
			_childContainer.RegisterType(Of IWpfHost, PluginViewOfHost)(New ContainerControlledLifetimeManager)
			_childContainer.RegisterType(Of PluginProcessProxy)(New ContainerControlledLifetimeManager)

			_startupInfp = New PluginStartupInfo With {.FullAssemblyPath = CatalogEntry.AssemblyPath, .AssemblyName = Path.GetFileNameWithoutExtension(CatalogEntry.AssemblyPath), .Bits = CatalogEntry.Bits, .MainClass = CatalogEntry.MainClass, .Name = CatalogEntry.Name, .Parameters = CatalogEntry.Parameters}

			_childContainer.RegisterInstance(_startupInfp)
		End Sub

		Private Sub OnProcessExited()
			If (Not _isDisposing) AndAlso (Not _fatalErrorOccurred) Then
				ReportError("Plugin process terminated unexpectedly", Nothing)
			End If
		End Sub

		Private Sub OnFatalError(ByVal ex As Exception)
			_fatalErrorOccurred = True
			ReportError(Nothing, ex)
		End Sub

		Private Sub ReportError(ByVal message As String, ByVal ex As Exception)
			RaiseEvent Error(Me, New PluginErrorEventArgs(Me, message, ex))
		End Sub
	End Class
End Namespace
