' Copyright (c) 2013 Ivan Krivyakov
' Licensed under the Apache License, Version 2.0 (the "License")
' you may not use this file except in compliance with the License.
' You may obtain a copy of the License at
' http://www.apache.org/licenses/LICENSE-2.0
'

Imports System.Collections.ObjectModel
Imports System.IO
Imports System.Reflection
Imports System.Threading.Tasks
Imports IKriv.WpfHost.Catalogs
Imports IKriv.WpfHost.Interfaces
Imports Microsoft.Practices.Unity

Namespace IKriv.WpfHost
	Friend Class PluginController
		Implements IDisposable

		Private ReadOnly _container As IUnityContainer
		Private ReadOnly _scheduler As TaskScheduler
		Private ReadOnly _errorHandlingService As ErrorHandlingService

        Private PrivateAvailablePlugins As IEnumerable(Of PluginCatalogEntry)
		Public Property AvailablePlugins As IEnumerable(Of PluginCatalogEntry)
			Get
                Return PrivateAvailablePlugins
			End Get
			Private Set(ByVal value As IEnumerable(Of PluginCatalogEntry))
                PrivateAvailablePlugins = value
			End Set
		End Property

        Private PrivateLoadedPlugins As ObservableCollection(Of Plugin)
		Public Property LoadedPlugins As ObservableCollection(Of Plugin)
			Get
                Return PrivateLoadedPlugins
			End Get
			Private Set(ByVal value As ObservableCollection(Of Plugin))
                PrivateLoadedPlugins = value
			End Set
		End Property

		Public Sub New(ByVal container As IUnityContainer, ByVal errorHandlingService As ErrorHandlingService)
			_container = container
			_errorHandlingService = errorHandlingService
			_scheduler = TaskScheduler.FromCurrentSynchronizationContext
			LoadedPlugins = New ObservableCollection(Of Plugin)
		End Sub

		Public Function LoadCatalogAcync() As Task
			Return Task.Factory.StartNew(AddressOf LoadCatalog)
		End Function

		Public Function LoadPluginAsync(ByVal info As PluginCatalogEntry) As Task(Of Plugin)
			Return Task.Factory.StartNew(Function() LoadPlugin(info)).ContinueWith(Function(loadPluginTask)
				Dim plugin = loadPluginTask.Result
				plugin.CreateView()
				LoadedPlugins.Add(plugin)
				Return plugin
			End Function, _scheduler)
		End Function

		Public Sub Dispose() Implements IDisposable.Dispose
			For Each plugin In LoadedPlugins
				DisposePlugin(plugin)
			Next plugin
		End Sub

		Public Sub RemovePlugin(ByVal plugin As Plugin)
			LoadedPlugins.Remove(plugin)
			DisposePlugin(plugin)
		End Sub

        Public Function GetUnsavedItems() As IDictionary(Of Plugin, String)
            Return LoadedPlugins.AsParallel.Select(Function(plugin) New With {Key .Plugin = plugin, Key .Data = GetUnsavedItems(plugin)}).Where(Function(item) item.Data IsNot Nothing).ToDictionary(Function(item) item.Plugin, Function(item) item.Data)
        End Function

		Public Function GetUnsavedItems(ByVal plugin As Plugin) As String()
			Try
				Dim service = plugin.GetService(Of IUnsavedData)()
				If service Is Nothing Then
					Return Nothing
				End If
				Return service.GetNamesOfUnsavedItems
			Catch ex As Exception
				_errorHandlingService.LogError("Error getting unsaved data from " & plugin.Title, ex)
				Return Nothing
			End Try
		End Function

		Private Sub DisposePlugin(ByVal plugin As Plugin)
			If plugin Is Nothing Then
				Exit Sub
			End If
			RemoveHandler plugin.Error, AddressOf OnPluginError

			Try
				plugin.Dispose()
			Catch ex As Exception
				_errorHandlingService.LogError("Error disposing plugin " & plugin.Title, ex)
			End Try
		End Sub

		Private Function LoadPlugin(ByVal info As PluginCatalogEntry) As Plugin
			Dim plugin = _container.Resolve(Of Plugin)()
			AddHandler plugin.Error, AddressOf OnPluginError

			Try
				plugin.Load(info)
			Catch e1 As Exception
				DisposePlugin(plugin)
				Throw
			End Try

			Return plugin
		End Function

		Private Sub LoadCatalog()
			Dim location = System.Reflection.Assembly.GetExecutingAssembly.Location
			Dim catalogFile = Path.Combine(Path.GetDirectoryName(location), "plugins.xml")
			Dim catalog = New XmlPluginCatalog(catalogFile)
			AvailablePlugins = catalog.GetPluginList
		End Sub

		Private Sub OnPluginError(ByVal sender As Object, ByVal args As PluginErrorEventArgs)
			Dim task = New Task(Sub() PluginErrorHandler(args))
			task.Start(_scheduler)
		End Sub

		Private Sub PluginErrorHandler(ByVal args As PluginErrorEventArgs)
			If args Is Nothing Then
				Exit Sub
			End If
			If args.Plugin Is Nothing Then
				Exit Sub
			End If

			Dim title As String = args.Plugin.Title
			Dim message As String = String.Format("An error occurred in plugin {0}. The plugin tab will be now closed." & vbCrLf & "{1}" & vbCrLf, title, args.Message)

			If LoadedPlugins.Contains(args.Plugin) Then
				_errorHandlingService.ShowError(message, args.Exception)
				LoadedPlugins.Remove(args.Plugin)
			Else
				_errorHandlingService.LogError(message, args.Exception)
			End If

			DisposePlugin(args.Plugin)
		End Sub
	End Class
End Namespace
