' Copyright (c) 2013 Ivan Krivyakov
' Licensed under the Apache License, Version 2.0 (the "License")
' you may not use this file except in compliance with the License.
' You may obtain a copy of the License at
' http://www.apache.org/licenses/LICENSE-2.0
'

Imports System.Configuration
Imports System.IO
Imports System.Threading
Imports IKriv.PluginHosting
Imports IKriv.WpfHost.Interfaces

Namespace IKriv.WpfHost
	Friend Class PluginProcessProxy
		Implements IDisposable

		Private ReadOnly _host As IWpfHost
		Private ReadOnly _startupInfo As PluginStartupInfo
		Private ReadOnly _errorHandlingService As ErrorHandlingService
		Private _readyEvent As EventWaitHandle
		Private _process As Process
		Private _name As String
		Private _pluginLoader As IPluginLoader

		Public ReadOnly Property Process As Process
			Get
				Return _process
			End Get
		End Property

        Private PrivateRemotePlugin As IRemotePlugin
		Public Property RemotePlugin As IRemotePlugin
			Get
                Return PrivateRemotePlugin
			End Get
			Private Set(ByVal value As IRemotePlugin)
                PrivateRemotePlugin = value
			End Set
		End Property

		Public Sub New(ByVal startupInfo As PluginStartupInfo, ByVal host As IWpfHost, ByVal errorHandlingService As ErrorHandlingService)
			_startupInfo = startupInfo
			_host = host
			_errorHandlingService = errorHandlingService
		End Sub

		Public Sub Start()
			If Process IsNot Nothing Then
				Throw New InvalidOperationException("Plugin process already started, cannot load more than one plugin per process")
			End If
			StartPluginProcess(_startupInfo.FullAssemblyPath)
		End Sub

		Public Sub LoadPlugin()
			If Process Is Nothing Then
				Throw New InvalidOperationException("Plugin process not started")
			End If
			If Process.HasExited Then
				Throw New InvalidOperationException("Plugin process has terminated unexpectedly")
			End If

			_pluginLoader = GetPluginLoader()
			RemotePlugin = _pluginLoader.LoadPlugin(_host, _startupInfo)
		End Sub

		Public Sub Dispose() Implements IDisposable.Dispose
			If RemotePlugin IsNot Nothing Then
				Try
					RemotePlugin.Dispose()
				Catch ex As Exception
					_errorHandlingService.LogError("Error disposing remote plugin for " & _startupInfo.Name, ex)
				End Try
			End If

			If _pluginLoader IsNot Nothing Then
				Try
					_pluginLoader.Dispose()
				Catch ex As Exception
					_errorHandlingService.LogError("Error disposing plugin loader for " & _startupInfo.Name, ex)
				End Try
			End If

			' this can take some time if we have many plugins; should be made asynchronous
			If Process IsNot Nothing Then
				Process.WaitForExit(5000)
				If Not Process.HasExited Then
					_errorHandlingService.LogError("Remote process for " & _startupInfo.Name & " did not exit within timeout period and will be terminated", Nothing)
					Process.Kill()
				End If
			End If
		End Sub

		Private Sub StartPluginProcess(ByVal assemblyPath As String)
            _name = "PluginProcess." & Guid.NewGuid.ToString
			Dim eventName = _name & ".Ready"
			_readyEvent = New EventWaitHandle(False, EventResetMode.ManualReset, eventName)

			Dim directory = Path.GetDirectoryName(Me.GetType.Assembly.Location)
			Dim exeFile = If(_startupInfo.Bits = 64, "PluginProcess64.exe", "PluginProcess.exe")
			Dim processName = Path.Combine(directory, exeFile)

			If Not File.Exists(processName) Then
				Throw New InvalidOperationException("Could not find file '" & processName & "'")
			End If

			Const quote As String = """"
			Const doubleQuote As String = """"""

			Dim quotedAssemblyPath = quote & assemblyPath.Replace(quote, doubleQuote) & quote
			Dim createNoWindow = Not Boolean.Parse(ConfigurationManager.AppSettings("PluginProcess.ShowConsole"))

			Dim info = New ProcessStartInfo With {.Arguments = _name & " " & quotedAssemblyPath, .CreateNoWindow = createNoWindow, .UseShellExecute = False, .FileName = processName}

			Trace.WriteLine(info.Arguments)

			_process = Process.Start(info)
		End Sub

		Private Function GetPluginLoader() As IPluginLoader
			If Process.HasExited Then
				Throw New InvalidOperationException("Plugin process has terminated unexpectedly")
			End If

			Dim timeoutMs = Int32.Parse(ConfigurationManager.AppSettings("PluginProcess.ReadyTimeoutMs"))

			If Not _readyEvent.WaitOne(timeoutMs) Then
				Throw New InvalidOperationException("Plugin process did not respond within timeout period")
			End If

			Dim hostChannelName = "WpfHost." & Process.GetCurrentProcess.Id
			IpcServices.RegisterChannel(hostChannelName)

			Dim url = "ipc://" & _name & "/PluginLoader"
			Dim pluginLoader = CType(Activator.GetObject(GetType(IPluginLoader), url), IPluginLoader)
			Return pluginLoader
		End Function
	End Class
End Namespace
